(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gont.ml,v 1.26 2001/11/27 03:51:56 malekith Exp $
 *
 * Module implementation: Gont
 * Description: Toplevel module for Gont compiler. There are two main entry
 * points here. First one, 'compile' compiles specifed source file, the 
 * second 'compile_iface' compiles interface. 
 *
 * Compilatation of source file involves:
 *  - initialization of Ksi codegenerator
 *  - opening input and output files
 *  - parsing it -- we parse one declaration at a time, and generate code
 *    for it. This generally involves sequence of:
 *      - Gontbind.bind -- binds names to symbols, reports any 
 *        undefined symbols
 *      - Gonttypes.typecheck_decl -- checks whatever declaration is 
 *        typecorrect
 *      - Gontcodegen.ksi_code -- returns list of Ksi declarations to
 *        be output (we might generate more then one Ksi declarations
 *        from given Gont declaration, hence list).
 *  - closing input and output files
 * 
 * Compilation of interface, is at the very moment simplified. It does
 * not check whatever given declaration has already been seen (we should
 * report error then), and few other things. We generally:
 *  - open input file
 *  - parse it, this results in list of declarations
 *  - pass it to Gontmodule.save function, that makes use of OCaml 
 *    serialization code.
 *  - close input file
 *)

type compilation_info = {ci_input_file: string;
                         ci_output_file: string;
                         ci_target_prefix: string;
                         ci_module_name: string}
open Gontutil
 
let open_files info =
  let in_ch = 
    if info.ci_input_file = "-" then 
       stdin 
     else 
       open_in info.ci_input_file in
  let out_ch = 
    if info.ci_output_file = "-" then 
       stdout
     else 
       open_out info.ci_output_file in
  (in_ch, out_ch)
let close_files (in_ch, out_ch) =
     if out_ch != stdout then close_out out_ch else ();
     if in_ch != stdin then close_in in_ch else ()
let loc (f, l, c) = {loc_file = f; loc_line = l; loc_char = c}
;;

let compile info =
  let (in_ch, out_ch) = open_files info in
  try
    Gontutil.reset ();
    Gontmodule.reset ();
    let bind_ctx = ref (Gontbind.init_bind info.ci_module_name) in
    let codegen_ctx = ref (Gontcodegen.init_codegen info.ci_module_name
                                                    info.ci_input_file) in
    let compile_decl decl =
      bind_ctx := Gontbind.bind !bind_ctx decl;
      Gonttypes.typecheck_decl decl;
      let (ctx', code) = Gontcodegen.ksi_code !codegen_ctx decl in
      codegen_ctx := ctx'; code in
    let lexbuf = Lexing.from_channel in_ch in
    Ksi.read_target_conf info.ci_target_prefix;
    Ksi.output_header out_ch;
    List.iter (Ksi.output_decl out_ch) 
              (Gontcodegen.ksi_runtime_proto !codegen_ctx);
    Lexer.set_file_name info.ci_input_file;
    (try
      while true do
        let decl = 
          try 
            (Parser.decl Lexer.token lexbuf) 
          with 
              Parsing.Parse_error ->
                raise(Syntax_error (loc (Lexer.get_current_location ()),
                                    "parse error"))
            | Lexer.Eof -> 
                raise Eof
            | Lexer.Error(f, l, c, e) ->
                raise(Syntax_error (loc (f, l, c), e))
        in
        List.iter (Ksi.output_decl out_ch) (compile_decl decl)
      done
    with Eof -> ());
    Gontcodegen.check_module !codegen_ctx;
    Ksi.output_trailer out_ch;
    close_files (in_ch, out_ch);
    0
  with 
      Syntax_error(loc, msg) ->
        close_files (in_ch, out_ch);
        err_msg (loc, msg);
        1
    | Error(msg) ->
        close_files (in_ch, out_ch);
        err_msg ({loc_file=info.ci_input_file; loc_line=0; loc_char=0}, msg);
        1
    | x ->
        close_files (in_ch, out_ch);
        prerr_endline ("Unhandled exception: " ^ (Printexc.to_string x));
        1
;;

let compile_iface info =
  let (in_ch, out_ch) = open_files info in
  let decls : decl list ref = ref [] in
  try 
    try
      Gontutil.reset ();
      Gontmodule.reset ();
      let lexbuf = Lexing.from_channel in_ch in
      Lexer.set_file_name info.ci_input_file;
      while true do
        let decl = 
          try 
            (Parser.iface_decl Lexer.token lexbuf) 
          with 
              Parsing.Parse_error ->
                raise(Syntax_error (loc (Lexer.get_current_location ()),
                                    "parse error"))
            | Lexer.Eof -> 
                raise Eof
            | Lexer.Error(f, l, c, e) ->
                raise(Syntax_error (loc (f, l, c), e))
        in
        decls := decl :: !decls
      done;
      assert false
    with Eof -> Gontmodule.save info.ci_module_name out_ch !decls;
                raise Eof
  with
      Eof ->
        close_files (in_ch, out_ch);
        0
    | Error(msg) ->
        close_files (in_ch, out_ch);
        err_msg ({loc_file=info.ci_input_file; loc_line=0; loc_char=0}, msg);
        1
    | Syntax_error(loc, msg) ->
        close_files (in_ch, out_ch);
        err_msg (loc, msg);
        1
    | x ->
        close_files (in_ch, out_ch);
        prerr_endline ("Unhandled exception: " ^ (Printexc.to_string x));
        1
;;

let dump_module m =
  print_endline (xf "// module dump of %s" m);
  let (g, t) = Gontmodule.globals m in
  List.iter (fun (t, n) -> 
                print_endline (xf "%s %s;" (Gonttypes.string_of_type t) n))
           g;
  List.iter (fun t -> 
                print_endline (xf "type %s = %s;" 
                                  (Gonttypes.string_of_type t.td_type) 
                                  t.td_name)) 
           t;
  0
;;
