(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontbind.ml,v 1.19 2001/11/27 03:51:56 malekith Exp $
 *
 * Module implementation: Gontbind
 * Description:
 * The purpouse of this file to make other part of compiler 
 * symbol-table-free. All occurences of Id(string) are changed
 * to Ref(symbol, int), where symbol is object type used for
 * symbols and int tells how deep in closure stack should we go. 
 *)

open Gontutil


(* General type of symbol object is defined in Gontutil, this
 * class just conforms it.
 *
 * Level stuff:
 *   fnc_level is level of function nesting for symbol.
 *     -2 for undefined symbols
 *     -1 for symbols from other modules
 *      0 for global symbols (in current module)
 *      1 for variables and parmaters to 1st level functions
 *      2 ditto for 2nd level...
 *   scope_level is level of scope nesting of symbol, levels -2 up to
 *   0 are the same as for fnc_level, next define level of block nesting.
 *
 * Example            // (fnc_level, scope_level)
 * int foo;           // (0, 0)
 * int bar(           // (0, 0)
 *     int x) {       // (1, 1)
 *     int baz;       // (1, 1)
 *     int f() {      // (1, 1)
 *         int k;     // (2, 2)
 *     }
 *     { int z; }     // (1, 2)
 *  }
 *)
 
class symbol = fun (scope, fnc, t, n) ->
  object
    val sym_type : ty = t
    val sym_name : id = n
    val sym_scope_level : int = scope
    val sym_fnc_level : int = fnc
    method get_name = sym_name
    method get_type = sym_type
    method get_fnc_level = sym_fnc_level
    method get_scope_level = sym_scope_level
    val mutable sym_closured = false
    method in_closure = sym_closured
    method put_into_closure = sym_closured <- true
  end

type ctx = {ctx_types: typedef list;
            ctx_syms: symbol list;
            ctx_fnc_level: int;
            ctx_scope_level: int;
            ctx_opened_modules: string list;
            ctx_module: string}

let make_symbol c t n =
  new symbol (c.ctx_scope_level, c.ctx_fnc_level, t, n)
;;

let init_bind modname = {ctx_syms = []; 
                         ctx_types = [];
                         ctx_fnc_level = 0; 
                         ctx_scope_level = 0;
                         ctx_opened_modules = [];
                         ctx_module = modname}

let find_mod_sym m n =
   try
     Gontmodule.import_symbol m n
   with 
     Gontmodule.Has_no_symbol ty -> 
        let sym = new symbol (-1, -1, ty, (m ^ "::" ^ n)) in
        Gontmodule.assign_symbol m n sym; sym

let find_mod_type m n =
     Gontmodule.import_type_symbol m n
;;

let find_sym ctx n =
  let qn = (ctx.ctx_module ^ "::" ^ n) in
  try
    List.find (fun sym -> sym#get_name = n || sym#get_name = qn) ctx.ctx_syms
  with Not_found -> 
    let m =
      List.find (fun m -> 
          try
            ignore(find_mod_sym m n);
            true
          with Not_found -> false) ctx.ctx_opened_modules in
      find_mod_sym m n

let find_type ctx m n =
  match m with
      "" -> 
        (try
          List.find (fun {td_name = x} -> n = x) ctx.ctx_types
         with Not_found ->
           let m = 
             List.find (fun m -> 
               try 
                 ignore(find_mod_type m n); 
                 true 
               with Not_found -> false) ctx.ctx_opened_modules in
           find_mod_type m n)
    | _ -> find_mod_type m n
;;

let prepare_ctx loc (c, parms) =
  let rec helper (c, parms) =
    match parms with
        ({parm_type=t; parm_name=n} as x)::xs -> 
          (let (fnc_lev, scope_lev) =
            try
              let sym = find_sym c n in
                (sym#get_fnc_level, sym#get_scope_level)
            with Not_found -> (-2, -2) in
          if fnc_lev == c.ctx_fnc_level then
            err (loc, xf "parameter `%s' defined twice" n);
          if scope_lev >= 0 then
            (* Note, that we don't warn about shodowing symbols
             * from other modules, that has been opened. *)
            warn (loc, xf "`%s' shadows other symbol" n));
          let sym = make_symbol c t n in
          x.parm_symbol <- Some sym;
          helper ({c with ctx_syms = sym :: c.ctx_syms }, xs)
      | [] -> c in
  helper ({c with ctx_fnc_level   = c.ctx_fnc_level + 1;
                  ctx_scope_level = c.ctx_scope_level + 1}, 
          parms)
;;

(* This does inplace modification of s. Updated context is returned. *)
let rec bind_stmt (c, s) =
  match s.raw_stmt with
      Skip -> c
    | Exp(e) -> bind_exp (c, e)
    | VarDef v ->
          (match v.var_symbol with
              None -> ()
            | Some _ -> assert false);
          let c = bind_type (c, v.var_type) in
          (let scope_lev =
            try
              let sym = find_sym c v.var_name in
                sym#get_scope_level
            with Not_found -> -2 in
          if scope_lev == c.ctx_scope_level then
            err (s.stmt_loc, xf "variable `%s' defined twice" v.var_name);
          if scope_lev >= 0 then
            warn (s.stmt_loc, xf "`%s' shodows other symbol" v.var_name));
          let sym = make_symbol c v.var_type v.var_name in
          v.var_symbol <- Some sym;
          {c with ctx_syms = sym :: c.ctx_syms}
    | Return(e) -> bind_exp (c, e)
    | ReturnVoid -> c
    | Block(ss) ->
        let rec bind_block (c, ss) =
          match ss with
              x::xs -> 
                let c' = bind_stmt (c, x) in
                bind_block (c', xs)
            | [] -> c in
        let new_scope = {c with ctx_scope_level = c.ctx_scope_level + 1} in
        let c' = bind_block(new_scope, ss) in
        ignore(c'); 
        c
    | Pair(s1, s2) ->
        bind_stmt ((bind_stmt (c, s1)), s2)
    | BreakIfNot e ->
        bind_exp (c, e)
    | Loop s ->
        bind_stmt (c, s)
    | If(e, s1, s2) ->
        bind_stmt (bind_stmt (bind_exp (c, e), s1), s2)
    | Open m ->
        Gontmodule.open_module m;
        {c with ctx_opened_modules = m :: c.ctx_opened_modules}
and bind_exp (c, e) =
  match e.raw_exp with
      SimpleBinOp(op, e1, e2) -> 
        let c' = bind_exp (c, e1) in
        bind_exp (c', e2)
    | SimpleUnOp(op, e1) ->
        bind_exp (c, e1)
    | Ref(r) ->
        (match r.ref_symbol with
            None ->
              let sym = 
                try
                  find_sym c r.ref_name
                with Not_found ->
                  err (e.exp_loc, xf "unbound symbol `%s'" r.ref_name) in
              r.ref_symbol <- Some sym;
              let lev = c.ctx_fnc_level - sym#get_fnc_level in
                assert (lev >= 0);
                r.ref_level <- if sym#get_fnc_level == 0 then 0 else lev
          (* We have already been here! *)
          | Some _ -> assert false); 
          (* Context is not updated in any way. *)
          c
    | CrossModuleRef(r) ->
        (match r.xmr_symbol with
            None ->
              (try
                 let sym = find_mod_sym r.xmr_module r.xmr_name in
                 r.xmr_symbol <- Some sym
               with Not_found ->
                 err (e.exp_loc, xf "there is no symbol %s in module %s"
                                    r.xmr_name r.xmr_module))
          | Some _ -> assert false);
          c
    | Call(e1, es) -> 
        let c' = bind_exp (c, e1) in
        let rec bind_parms (c, pp) =
            match pp with
                p::ps -> 
                  let c' = bind_exp (c, p) in
                  bind_parms (c', ps)
              | [] -> c in
        bind_parms (c', es)
    | IntConst(_) -> c
    | BoolConst(_) -> c
    | FloatConst(_) -> c
    | StringConst(_) -> c
    | Lambda {lambda_parms=p; lambda_ret_type=r; lambda_body=s} ->
       let new_ctx = prepare_ctx e.exp_loc (c, p) in
       let c' = bind_type (new_ctx, make_fn_type (r, p)) in
       let c'' = bind_stmt (c', s) in
       ignore(c'');
       (* We return the same contex. Local function cannot affect
        * parent enviroment. *)
       c
and bind_type (c, t) =
  let f t =
    match t with
        NamedType ({nt_tydef = None} as r) ->
          (try
             r.nt_tydef <- Some (find_type c r.nt_module r.nt_name);
             NamedType r
           with Not_found ->
             if r.nt_module = "" then
               err (r.nt_loc, xf "use of unknown type %s" r.nt_name)
             else
               err (r.nt_loc, xf "module %s does not define type %s"
                                 r.nt_module r.nt_name))
      | NamedType _ -> assert false
      | Int | Void | Bool | Float | String | TyVar _ | AbstractType
      | Tuple _ | Func _ | Struct _ | OptStruct _ | Union _ -> t in
  ignore(type_walk f t);
  c
;;

(* Modify stmts in place, adding symbol entries to Ref terms. This musn't be
 * called more then once in a toplevel function. *)
let bind c decl =
  let add_global loc t n =
    try
      ignore(find_sym c n);
      err (loc, xf "redefintion of global symbol `%s'" n)
    with Not_found -> make_symbol c t n in
  match decl with
      FuncDecl f ->
        let fntype = make_fn_type (f.fn_ret_type, f.fn_parms) in
        let c = bind_type (c, fntype) in
        let sym = add_global f.fn_loc
                             fntype
                             (c.ctx_module ^ "::" ^ f.fn_name) in
        let c' = {c with ctx_syms = sym :: c.ctx_syms} in
        f.fn_symbol <- Some sym;
        let initial_ctx = prepare_ctx f.fn_loc (c', f.fn_parms) in
        ignore (bind_stmt (initial_ctx, f.fn_body));
        c'
    | GlobalDecl g ->
        let c = bind_type (c, g.global_type) in
        let sym = add_global g.global_loc 
                             g.global_type 
                             (c.ctx_module ^ "::" ^ g.global_name) in
        g.global_symbol <- Some sym;
        {c with ctx_syms = sym :: c.ctx_syms}
    | ExternC x ->
        let c = bind_type (c, x.xc_type) in
        let sym = add_global x.xc_loc x.xc_type x.xc_name in
        x.xc_symbol <- Some sym;
        {c with ctx_syms = sym :: c.ctx_syms}
    | TopLevelOpen m ->
        Gontmodule.open_module m;
        {c with ctx_opened_modules = m :: c.ctx_opened_modules}
    | TypeDef ({td_name = n} as t) ->
        (try 
          let s = find_type c "" n in
          if s.td_module = "" then
            err (t.td_loc, xf "redefintion of type `%s'" n)
          else
            warn (t.td_loc, xf "type `%s' shadows `%s::%s'" 
                               n s.td_module s.td_name)
         with Not_found -> ());
        let c = bind_type (c, t.td_type) in
        {c with ctx_types = t :: c.ctx_types}
    | FuncProto {fp_loc = loc} -> 
        err (loc, "sorry, prototypes unimplemented")
;;

(* Apply given function, of (exp -> 'a list) type, for all
 * expressions in a tree, returning flat list of results. *)
let rec for_all_exp exp_fnc s =
  let stmt_fnc s = for_all_exp exp_fnc s in
  match s.raw_stmt with
      Skip -> []
    | Exp(e) -> exp_fnc e
    | Return(e) -> exp_fnc e
    | ReturnVoid -> []
    | Block(l) -> (List.flatten (List.map stmt_fnc l))
    | Pair(a, b) -> (stmt_fnc a) @ (stmt_fnc b)
    | VarDef(_) -> []
    | BreakIfNot e -> exp_fnc e
    | Loop s -> stmt_fnc s
    | If(e, s1, s2) -> exp_fnc e @ stmt_fnc s1 @ stmt_fnc s2
    | Open _ -> []
;;

(* Return list of symbols, at given scoping level, that needs to put into the
 * closure becouse they are referenced in stmts. *)
let closure lev s =
  let need_closure = ref false in
  let rec exp_closure lambda_lev e =
    let ec = exp_closure lambda_lev in
    match e.raw_exp with
        SimpleBinOp(op, e1, e2) -> (ec e1) @ (ec e2)
      | SimpleUnOp(op, e1) -> ec e1
      | Ref(r) ->
          (match r.ref_symbol with
              None -> assert false
            | Some sym ->
                if r.ref_level > 0 then
                  (if lambda_lev > 0 then need_closure := true;
                   if sym#get_fnc_level = lev && not sym#in_closure then
                     (sym#put_into_closure; [sym])
                   else
                     [])
                else
                  [])
      | Call(e1, es) ->
          (ec e1) @ (List.flatten (List.map ec es))
      | CrossModuleRef _ -> []
      | IntConst(_) -> []
      | BoolConst(_) -> []
      | FloatConst(_) -> []
      | StringConst(_) -> []
      | Lambda {lambda_body=s} ->
         for_all_exp (exp_closure (lambda_lev + 1)) s in
  (!need_closure, for_all_exp (exp_closure 0) s)
;;

